// Author: Alberto González Palomo http://sentido-labs.com
// ©2017 Alberto González Palomo http://sentido-labs.com
// Created: 2017-02-25 18:47
var model = { query: "", reverse: false, url: "jpn-eng.tei" };

var nexus = dom.Nexus.new(model);

var dict = [];
var dictIndex = { direct:[], inverse:[] };

// 1. Load the XML file and build the entry index for quick search.
// 2. Add the search box and checkbox, bind them to the model.
// 3. Listen for changes in the model, filter results and build HTML view.

// Helper function to display a brief message to the user:
function status(message) { dom.id('results').textContent = message; }

// Helper function to load XML files:
function getXML(url, progress, finished)
{
    var xhr = new XMLHttpRequest();
    xhr.addEventListener('error', function (event) {
        finished(event.target.status + ', ' +
                 (event.target.statusText || 'can not load ' + url));
    });
    xhr.addEventListener('load', function (event) {
        var xml = event.target.responseXML;
        if (!xml) {
            try {
                var parser = new DOMParser();
                xml = parser.parseFromString(event.target.responseText,
                                             'application/xml');
            } catch (e) {
                finished(e);
                return;
            }
        }
        finished(null, xml);
    });
    if (progress) xhr.addEventListener('progress', progress);
    xhr.open('get', url);
    xhr.send();
}

getXML(model.url, function progress(event) {
    var units = 'Bytes';
    var total = event.total;
    ['KB', 'MB', 'GB'].some(function (unitSymbol) {
        if (total < 1000) return true;
        total /= 1000;
        units = unitSymbol;
        return false;
    });
    status('Loading ' + model.url + ', ' +
           Math.round(event.loaded * 100 / event.total) + '% of ' +
           total.toFixed(2) + ' ' + units);
}, function finished(err, dictXML) {
    if (err) {
        console.error(err);
        alert('Error: ' + err);
        status('You can download a dictionary from ' +
               'https://github.com/freedict/fd-dictionaries/raw/master/jpn-deu/jpn-deu.tei');
        return;
    }

    var xpath = dom.XPath.new(dictXML, {
        "tei": "http://www.tei-c.org/ns/1.0"
    });

    var title = xpath('tei:TEI/tei:teiHeader//tei:title[1]').str(dictXML);
    if (title) {
        document.title = title;
        dom([ dom.clear(dom.$('header > h1')), title ]);
    }

    var iter = xpath('tei:TEI/tei:text/tei:body/tei:entry').iter(dictXML);
    var formXPath = xpath('tei:form');
    var orthXPath = xpath('tei:orth');
    var pronXPath = xpath('tei:pron');
    var defXPath  = xpath('tei:sense/tei:cit/tei:quote|tei:sense/tei:def');
    dom.iterateAsync(iter, function forEach(entryElement) {
        var iter, element;
        var entry = { "forms":[], "senses": [] };

        iter = formXPath.iter(entryElement);
        while ((element = iter.iterateNext())) {
            entry.forms.push({
                "orth": orthXPath.str(element),
                "pron": pronXPath.str(element)
            });
        }

        iter = defXPath.iter(entryElement);
        while ((element = iter.iterateNext())) {
            entry.senses.push(element.textContent);
        }

        dict.push(entry);
    }, function blockFinished() {
        status('Read ' + dict.length + ' entries.');
    }, function finished() {
        status('Indexing ' + dict.length + ' entries.');
        // Let things settle for a bit and detach buildIndex() from this
        // closure so that dictXML etc. can be garbage-collected before
        // we start building the index.
        setTimeout(buildIndex, 260);
    });
});

function buildIndex() {
    var iter = dom.iterFromArray(dict);
    dom.iterateAsync(iter, function forEach(entry) {
        var text;
        text = entry.forms.map(function (form) {
            var text = form.orth.toLowerCase();
            if (form.pron) text += ' ' + form.pron.toLowerCase();
            return text;
        }).join(' ');
        dictIndex.direct.push(text);
        text = entry.senses.map(function (sense) {
            return sense.toLowerCase();
        }).join(' ');
        dictIndex.inverse.push(text);
    }, function blockFinished() {
        status(dictIndex.direct.length + ' entries indexed.');
    }, function finished() {
        // Trigger first view update from values in model.query:
        nexus.changed('query');
    });
}


// Define a couple of UI components to encapsulate text fields and checkboxes:
var Input = {
    new: function Input(property, nexus) {
        return this.use(dom(["input", { "name":property } ]), nexus);
    },
    use: function Input(view, nexus) {
        var property = view.name;
        return nexus.bind(this, property, 'input', view);
    },
    get: function get(view) { return view.value.trim(); },
    set: function set(view, value) { view.value = value.trim(); }
};

var CheckBox = {
    new: function CheckBox(property, nexus) {
        return this.use(dom(["input", {
            "name": property, "type": "checkbox"
        }]), nexus);
    },
    use: function CheckBox(view, nexus) {
        var property = view.name;
        return nexus.bind(this, property, 'change', view);
    },
    get: function get(view) { return view.checked; },
    set: function set(view, value) { view.checked = value; }
};

// Here we get the document title from the <title> tag, and replace the
// contents of the <h1> tag inside the <header> with it:
dom([ dom.clear(dom.$('header > h1')), document.title ]);

// Activate the query UI:
Input.use(dom.$('input[name=query]'), nexus);
CheckBox.use(dom.$('input[name=reverse]'), nexus);

// Add a timestamp:
dom([
    dom.$("footer"),
    "Last updated: ", (new Date()).toISOString().replace(/T.*/, '')
]);

function scoreFor(query, entry)
{
    var score = 0.0;
    if ('string' === typeof entry) {
        entry = entry.toLowerCase();
        if        (entry.startsWith(query)) {
            score = 4.0 + query.length/entry.length;
        } else if (entry.endsWith(query)) {
            score = 3.0 + query.length/entry.length;
        } else if (entry.indexOf(query) >= 0) {
            score = 3.0 *(query.length/entry.length);
        }
    } else if (Array.isArray(entry)) {
        score = entry.reduce(function (score, item) {
            return Math.max(score, scoreFor(query, item));
        }, score);
    } else {
        Object.keys(entry).forEach(function (key) {
            score = Math.max(score, scoreFor(query, entry[key]));
        });
    }
    return score;
}

// Watch changes in the model and display the results:
nexus.addListener('*', function (model, name, origin) {
    dom([dom.clear(dom.id('results')), ["progress"]]);
    var value = model.query.toLowerCase();
    var results = [];
    if (value) {
        // Still find matches in the other direction,
        // but give them lower score.
        var indexes = model.reverse?
            [dictIndex.inverse, dictIndex.direct]:
            [dictIndex.direct, dictIndex.inverse];
        indexes.forEach(function (searchIndex, rank) {
            searchIndex.forEach(function (text, i) {
                if (text.indexOf(value) >= 0) {
                    var entry = dict[i];
                    results.push({
                        'score': scoreFor(value, entry) / (rank+1),
                        'entry': entry
                    });
                }
            });
        });
    }

    // Pick the best 20:
    results.sort(function (a, b) { return b.score - a.score; });
    results.splice(20);

    dom([dom.clear(dom.id('results')),
         results.reduce(function (a, result) {
             var scoreView = [
                 "div.score",
                 ['span.score-number', result.score.toFixed(1)],
                 ['div.score-gauge-container',
                  ['div.score-gauge', {
                      "style": "width:" + Math.round(result.score/0.05) + "%"
                  }]
                 ]
             ];
             a.push(['div.entry',
                     scoreView,
                     result.entry.forms.reduce(function (a, form) {
                         if (a.length > 1) a.push(', ');
                         a.push(['b', form.orth]);
                         if (form.pron) {
                             a.push(' ');
                             a.push(['i', form.pron]);
                         }
                         return a;
                     }, ['span.forms']),
                     ' ',
                     result.entry.senses.reduce(function (a, sense) {
                         if (a.length > 1) a.push(', ');
                         a.push(['span', sense]);
                         return a;
                     }, ['span.senses'])
                    ]);
             return a;
         }, [''])
        ]);
    if (results.length < 1) {
        var message = value?
            'There are no words that contain "' + value + '".':
            'Ready.';
        dom([dom.id('results'), message]);
    }
});
