// Example of using Bulma.io CSS with Domo JS.

var model = {
    "state": "",
    "autocycle": "",
    "loading": "",
    "text": "Temporary notification content.",
    "otherText": "This other text is for permanent notifications.",
    "progress": 0,
    "notificationTimeout": 3678
};

// Nexus handles the events bound to model properties, and change notifications.
var nexus = dom.Nexus.new(model);


// Buttons can have different states, which are indicated by CSS class names.
// In this case, Bulma's classes are is-primary etc.
// Here we have three cycles: the first one is just a show of different
// button/notification styles in Bulma.
// 1. <nothing> -> primary -> info -> success -> warning -> danger -> <nothing>
// 2. <nothing> -> active -> <nothing>
// 3. <nothing> -> loading -> <nothing>
var buttonStates = {
    "state": [
        "", "is-primary", "is-info", "is-success", "is-warning", "is-danger"
    ],
    "autocycle": "is-active",
    "loading":   "is-active"
};


// Now we activate some HTML elements with <component>.use(...)

dom.id('explanation-buttons-cycle')
    .appendChild(dom(JSON.stringify(buttonStates.state)));

dom.$$('button[name]').forEach(function (button) {
    var states = buttonStates[button.name];
    if (states) {
        Bulma.Button.use(button, { "click": states }, nexus);
    } else {
        Bulma.Button.use(button, null, nexus);
    }
});

dom.$$('input[name=text],input[name=otherText]').forEach(function (input) {
    Bulma.Input.use(input, nexus);
});

dom.iterateAsync(dom.iter('button.delete'), function forEach(button) {
    console.log('delete button is', button);
    Bulma.DeleteButton.use(button);
});

dom.$$('progress[name]').forEach(function (progress) {
    Bulma.Progress.use(progress, progress.getAttribute('name'), nexus);
});

dom.$$('input[name]').forEach(function (input) {
    Bulma.Input.new(input.name, nexus, input);
});


// Notifications, either temporary (disappear after a while) or permanent.
function notify(className, content, timeout)
{
    dom([dom.id('notifications'),
         Bulma.Notification.new(className, content, timeout)
        ]);
}

function timestamp() { return ['tt.timestamp', (new Date()).toTimeString()]; }

dom.$$('button[name=notify-temporarily]').forEach(function (button) {
    dom([button, {
        "click": function (event) {
            notify(model.state, model.text,
                   model.notificationTimeout)
            ++model.progress;
            nexus.changed('progress', this);
        }
    }]);
    Bulma.Button.new(null, null, nexus);
});

dom.$$('button[name=notify-permanently]').forEach(function (button) {
    dom([button, {
        "click": function (event) {
            notify(model.state, ['', timestamp(), ' ', model.otherText],
                   0)
            ++model.progress;
            nexus.changed('progress', this);
        }
    }]);
    Bulma.Button.use(button, null, nexus);
});

// Display the whole model any time a property changes:
nexus.addListener("*", function (model, property, origin) {
    dom.id("model").textContent = JSON.stringify(model, null, '  ');
});

// Now me modify a property in the model without user intervention,
// and UI elements will be updated accordingly.
var autocycler = (function (property, nexus, states, delay/*ms*/) {
    var interval;
    function cycle()
    {
        var state = nexus.model[property];
        var index = states.indexOf(state);
        if (index >= 0) {
            nexus.model[property] = states[(index + 1) % states.length];
            nexus.changed(property);
        }
    }
    return {
        start: function start() {
            if (!interval) interval = setInterval(cycle, 670);
        },
        stop: function stop() {
            if (interval) interval = clearInterval(interval);
        }
    };
})('state', nexus, buttonStates.state, 670);

// Activate/deactivate the autocycler.
nexus.addListener("autocycle", function (model, property, origin) {
    switch (model[property]) {
    case 'is-active': autocycler.start(); break;
    case '':          autocycler.stop();  break;
    default: alert('Unexpected ' + property + ' value: ' + model[property]);
    }
});

// Bulma's buttons can display a "loading" spinner, and here we bind that
// appearance to the value of the "loading" property in the model.
nexus.addListener("loading", function (model, property, origin) {
    var classList = dom.id('center').classList;
    var notifications = dom.id('notifications');
    switch (model[property]) {
    case 'is-active':
        classList.add('is-loading');
        notify('is-success', 'Now loading.',
               model.notificationTimeout);
        break;
    case '':
        classList.remove('is-loading');
        notify('is-info', 'The center button is not in the loading state.',
               model.notificationTimeout);
        break;
    default: alert('Unexpected ' + property + ' value: ' + model[property]);
    }
});
